import typer

from chroma_ops.rebuild_fts import rebuild_fts
from chroma_ops.wal_commit import command as commit_wal_command
from chroma_ops.wal_clean import command as clean_wal_command
from chroma_ops.wal_export import command as export_wal_command
from chroma_ops.info import command as info_command
from chroma_ops.clean import command as clean_command

app = typer.Typer(no_args_is_help=True, help="ChromaDB Ops Commands.")

app.command(
    name="commit-wal", help="Commit WAL to HNSW lib binary index", no_args_is_help=True
)(commit_wal_command)

app.command(
    name="clean-wal",
    help="Cleans up WAL and VACUUM the SQLite DB.",
    no_args_is_help=True,
)(clean_wal_command)
app.command(
    name="export-wal", help="Exports the WAL to a jsonl file.", no_args_is_help=True
)(export_wal_command)
app.command(
    name="rebuild-fts", help="Rebuilds Full Text Search index.", no_args_is_help=True
)(rebuild_fts)

app.command(
    name="info",
    help="Provide persistent Chroma DB information. "
    "Useful to understand how your Chroma works or get support from the team.",
    no_args_is_help=True,
)(info_command)

app.command(
    name="clean",
    help="Clean up orphaned vector segment directories.",
    no_args_is_help=True,
)(clean_command)

if __name__ == "__main__":
    app()



#!/usr/bin/env python3
import argparse
import os
import sqlite3
import sys
from typing import Optional, Sequence

import typer
from chromadb import __version__ as chroma_version
from chroma_ops.utils import (
    validate_chroma_persist_dir,
    get_hnsw_index_ids,
    get_dir_size,
    PersistentData,
    decode_seq_id,
)


def clean_wal(
    persist_dir: str,
    skip_collection_names: Optional[Sequence[str]] = None,
    database: Optional[str] = "default",
    namespace: Optional[str] = "default",
) -> None:
    validate_chroma_persist_dir(persist_dir)
    typer.echo(f"Size before: {get_dir_size(persist_dir)}", file=sys.stderr)
    sql_file = os.path.join(persist_dir, "chroma.sqlite3")
    conn = sqlite3.connect(sql_file)
    # conn = sqlite3.connect(f"file:{sql_file}?mode=ro", uri=True)

    cursor = conn.cursor()
    if int(chroma_version.split(".")[1]) <= 4:
        collection_name_index = 4
        query = "SELECT s.id as 'segment',s.topic as 'topic', c.id as 'collection' , c.dimension as 'dimension', c.name FROM segments s LEFT JOIN collections c ON s.collection = c.id WHERE s.scope = 'VECTOR';"
    else:
        collection_name_index = 3
        query = "SELECT s.id as 'segment', c.id as 'collection' , c.dimension as 'dimension', c.name FROM segments s LEFT JOIN collections c ON s.collection = c.id WHERE s.scope = 'VECTOR';"

    cursor.execute(query)

    results = cursor.fetchall()
    wal_cleanup_queries = []
    for row in results:
        if (
            skip_collection_names
            and row[collection_name_index] in skip_collection_names
        ):
            continue
        if int(chroma_version.split(".")[1]) <= 4:
            segment_id = row[0]
            topic = row[1]
            collection_id = row[2]
        else:
            segment_id = row[0]
            collection_id = row[1]
            topic = f"persistent://{database}/{namespace}/{collection_id}"
        metadata_pickle = os.path.join(persist_dir, segment_id, "index_metadata.pickle")
        if os.path.exists(metadata_pickle):
            metadata = PersistentData.load_from_file(metadata_pickle)
            if hasattr(metadata, "max_seq_id"):
                max_seq_id = metadata.max_seq_id
            else:
                max_seq_id_query_hnsw_057 = (
                    "SELECT seq_id FROM max_seq_id WHERE segment_id = ?"
                )
                cursor.execute(max_seq_id_query_hnsw_057, [row[0]])
                results = cursor.fetchall()
                max_seq_id = decode_seq_id(results[0][0]) if len(results) > 0 else 0
            wal_cleanup_queries.append(
                f"DELETE FROM embeddings_queue WHERE seq_id < {max_seq_id} AND topic='{topic}';"
            )
        else:
            hnsw_space = cursor.execute(
                "select str_value from collection_metadata where collection_id=? and key='hnsw:space'",
                (collection_id,),
            ).fetchone()
            if not hnsw_space:
                continue
            hnsw_space = "l2" if hnsw_space is None else hnsw_space[0]
            list_of_ids = get_hnsw_index_ids(
                f"{os.path.join(persist_dir, segment_id)}", hnsw_space, row[3]
            )
            batch_size = 100
            for batch in range(0, len(list_of_ids), batch_size):
                wal_cleanup_queries.append(
                    f"DELETE FROM embeddings_queue WHERE seq_id IN ({','.join([str(i) for i in list_of_ids[batch:batch + batch_size]])});"
                )
    if len(wal_cleanup_queries) > 0:
        typer.echo("Cleaning up WAL", file=sys.stderr)
        wal_cleanup_queries.append("VACUUM;")
        cursor.executescript("\n".join(wal_cleanup_queries))
    # Close the cursor and connection
    cursor.close()
    conn.close()
    typer.echo(f"Size after: {get_dir_size(persist_dir)}", file=sys.stderr)


def command(
    persist_dir: str = typer.Argument(..., help="The persist directory"),
    skip_collection_names: str = typer.Option(
        None,
        "--skip-collection-names",
        "-s",
        help="Comma separated list of collection names to skip",
    ),
) -> None:
    clean_wal(persist_dir, skip_collection_names=skip_collection_names)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("persist_dir", type=str)
    parser.add_argument("--skip-collection-names", type=str, default=None)
    arg = parser.parse_args()
    clean_wal(arg.persist_dir, skip_collection_names=arg.skip_collection_names)



import argparse
import os
import shutil
import sqlite3
import sys
import uuid

import typer
from chroma_ops.utils import validate_chroma_persist_dir


def clean(persist_dir: str):
    validate_chroma_persist_dir(persist_dir)
    sql_file = os.path.join(persist_dir, "chroma.sqlite3")
    conn = sqlite3.connect(f"file:{sql_file}?mode=ro", uri=True)
    cursor = conn.cursor()

    print("Cleaning up orphanated segment dirs...", file=sys.stderr)
    query = "SELECT id FROM segments WHERE scope = 'VECTOR';"
    cursor.execute(query)
    results = cursor.fetchall()
    active_segments = []
    for result in results:
        active_segments.append(result[0])
    cursor.close()
    conn.commit()
    conn.close()
    # list dirs in persist_dir
    for dir in os.listdir(persist_dir):
        if os.path.isdir(os.path.join(persist_dir, dir)) and dir not in active_segments and os.path.exists(os.path.join(persist_dir, dir, "header.bin")):
            print(f"Deleting orphanated segment dir: {dir}", file=sys.stderr)
            shutil.rmtree(os.path.join(persist_dir, dir))


def command(
    persist_dir: str = typer.Argument(..., help="The persist directory"),
) -> None:
    clean(persist_dir)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("persist_dir", type=str, help="The persist directory")
    arg = parser.parse_args()
    clean(arg.persist_dir)



#!/usr/bin/env python3
import argparse
import os
import sys
from typing import Sequence, Optional

import chromadb
import typer

from chromadb.segment.impl.vector.local_persistent_hnsw import PersistentData
from chromadb.segment import VectorReader
from chromadb.types import SegmentScope, Operation

from chroma_ops.utils import validate_chroma_persist_dir


def commit_wal(
    persist_dir: str, skip_collection_names: Optional[Sequence[str]] = None
) -> None:
    """Note this uses internal ChromaDB APIs which may change at any moment."""

    validate_chroma_persist_dir(persist_dir)
    # TODO add version check and confirm with the user before proceeding
    client = chromadb.PersistentClient(
        path=persist_dir
    )  # TODO we inadvetently migrate the targe DB to whatever version of Chroma is installed
    all_collections = client.list_collections()
    vector_segments = []
    for col in all_collections:
        if (
            col.name in skip_collection_names
            if skip_collection_names
            else [] or col.count() == 0
        ):
            typer.echo(f"Ignoring skipped collection {col.name}", file=sys.stderr)
            continue
        vector_segments.extend(
            [
                s
                for s in client._server._sysdb.get_segments(collection=col.id)
                if s["scope"] == SegmentScope.VECTOR
            ]
        )

    for s in vector_segments:
        col = client._server._get_collection(
            s["collection"]
        )  # load the collection and apply WAL
        client._server._manager.hint_use_collection(
            s["collection"], Operation.ADD
        )  # Add hint to load the index into memory
        segment = client._server._manager.get_segment(
            s["collection"], VectorReader
        )  # Get the segment instance
        segment._apply_batch(segment._curr_batch)  # Apply the current WAL batch

        segment._persist()  # persist the index
        PersistentData.load_from_file(
            os.path.join(persist_dir, str(s["id"]), "index_metadata.pickle")
        )  # load the metadata after persisting
        typer.echo(
            f"Processing index for collection {col['name']} ({s['id']}) - "
            f"total vectors in index {len(segment._index.get_ids_list())}",
            file=sys.stderr,
        )
        segment.close_persistent_index()


def command(
    persist_dir: str = typer.Argument(..., help="The persist directory"),
    skip_collection_names: str = typer.Option(
        None,
        "--skip-collection-names",
        "-s",
        help="Comma separated list of collection names to skip",
    ),
) -> None:
    commit_wal(persist_dir, skip_collection_names=skip_collection_names)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("persist-dir", type=str)
    parser.add_argument("--skip-collection-names", type=str, default=None)
    arg = parser.parse_args()
    commit_wal(arg.persist_dir, skip_collection_names=arg.skip_collection_names)



import os
import pickle
from typing import List, cast, Optional, Dict, Union

import hnswlib


def validate_chroma_persist_dir(persist_dir: str) -> None:
    if not os.path.exists(persist_dir):
        raise ValueError(f"Persist dir ({persist_dir}) does not exist")
    if not os.path.exists(f"{persist_dir}/chroma.sqlite3"):
        raise ValueError(
            f"{persist_dir} does not appear to be valid ChromaDB persist directory"
        )


def get_hnsw_index_ids(filename: str, space: str = "l2", dim: int = 384) -> List[int]:
    index = hnswlib.Index(space=space, dim=dim)
    index.load_index(
        filename,
        is_persistent_index=True,
        max_elements=100000,
    )
    ids = index.get_ids_list().copy()
    index.close_file_handles()
    return cast(List[int], ids)


def read_script(script: str) -> str:
    return open(
        os.path.join(os.path.dirname(os.path.realpath(__file__)), script), "r"
    ).read()


def get_dir_size(path: str) -> int:
    total_size = 0
    for dirpath, dirnames, filenames in os.walk(path):
        for f in filenames:
            fp = os.path.join(dirpath, f)
            if os.path.exists(fp):
                total_size += os.path.getsize(fp)
    return total_size


def get_file_size(path: str) -> int:
    # ensure it is a file
    if not os.path.isfile(path):
        raise ValueError(f"{path} is not a file")
    return os.path.getsize(path)


SeqId = int


# TODO this works for versions of Chroma using pickle persistent data.
class PersistentData:
    """Stores the data and metadata needed for a PersistentLocalHnswSegment"""

    dimensionality: Optional[int]
    total_elements_added: int
    max_seq_id: SeqId

    id_to_label: Dict[str, int]
    label_to_id: Dict[int, str]
    id_to_seq_id: Dict[str, SeqId]

    def __init__(
        self,
        dimensionality: Optional[int],
        total_elements_added: int,
        max_seq_id: int,
        id_to_label: Dict[str, int],
        label_to_id: Dict[int, str],
        id_to_seq_id: Dict[str, SeqId],
    ):
        self.dimensionality = dimensionality
        self.total_elements_added = total_elements_added
        self.max_seq_id = max_seq_id
        self.id_to_label = id_to_label
        self.label_to_id = label_to_id
        self.id_to_seq_id = id_to_seq_id

    @staticmethod
    def load_from_file(filename: str) -> "PersistentData":
        """Load persistent data from a file"""
        with open(filename, "rb") as f:
            ret = cast(PersistentData, pickle.load(f))
            return ret


def decode_seq_id(seq_id_bytes: Union[bytes, int]) -> SeqId:
    """Decode a byte array into a SeqID"""
    if isinstance(seq_id_bytes, int):
        return seq_id_bytes

    if len(seq_id_bytes) == 8:
        return int.from_bytes(seq_id_bytes, "big")
    elif len(seq_id_bytes) == 24:
        return int.from_bytes(seq_id_bytes, "big")
    else:
        raise ValueError(f"Unknown SeqID type with length {len(seq_id_bytes)}")


# https://stackoverflow.com/a/1094933
def sizeof_fmt(num: int, suffix: str = "B") -> str:
    n: float = float(num)
    for unit in ("", "Ki", "Mi", "Gi", "Ti", "Pi", "Ei", "Zi"):
        if abs(n) < 1024.0:
            return f"{n:3.1f}{unit}{suffix}"
        n /= 1024.0
    return f"{n:.1f}Yi{suffix}"



import argparse
import os
import sqlite3
import sys

import chromadb
import typer

from chroma_ops.utils import validate_chroma_persist_dir, read_script


def rebuild_fts(persist_dir: str) -> None:
    validate_chroma_persist_dir(persist_dir)
    sql_file = os.path.join(persist_dir, "chroma.sqlite3")
    conn = sqlite3.connect(sql_file)
    cursor = conn.cursor()
    script = read_script("scripts/drop_fts.sql")
    cursor.executescript(script)
    cursor.close()
    conn.close()
    typer.echo("Dropped FTS. Will try to start your Chroma now.", file=sys.stderr)
    typer.echo(
        "NOTE: Depending on the size of your documents in Chroma it may take a while for Chroma to start up again.",
        file=sys.stderr,
        color=typer.colors.YELLOW,
    )
    try:
        chromadb.PersistentClient(path=persist_dir)
        typer.echo("Chroma started successfully.", file=sys.stderr)
    except Exception as e:
        typer.echo(
            f"Chroma failed to start. Error: {repr(e)}",
            file=sys.stderr,
            color=typer.colors.RED,
            err=True,
        )


def command(
    persist_dir: str = typer.Argument(..., help="The persist directory"),
) -> None:
    rebuild_fts(persist_dir)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("persist_dir", type=str)
    arg = parser.parse_args()
    rebuild_fts(arg.persist_dir)
